//
// This file containes dual-rail flip-flops implementations built on
// timing assumptions. These flip-flops enforces all-zerous spacer on
// positive phase of each odd clock cycle and all-ones spacer on positive
// phase of each even clock cycle. 
//
// Pros:
// (+) Built-in controller decides which spacer to enforce without any new 
//     input to flip-flops.
//
// Cons:
// (-) As the same controller is built-in into each flip-flop, the size 
//     of the circuit can grow significantly comparing to non-alternaing 
//     spacer implemntation.
// (-) DF8, DF82, DF84, LOGIC0, LOGIC1 do not have reset input which is 
//     necessary to reset spalt_clk (indicates every second clock cycle).
//     The solution to this is to introduce reset input in  dual-rail 
//     implementations of the above modules.
//
//  Notes:
//  (!) Single-rail circuits containing DF8, DF82 or DF84 flip-flops (wisout
//      reset input) cannot be converted into dual-rail circuit with 
//      alternating spacer.
//


// Delay for flip-flops with timing assumptions

module delay (A, Q);
  input A;
  output Q;

  BUF1 inst_q (.A(A), .Q(Q));
endmodule


// Basic DF with  alternating spacer output protocol
// ( SP0 -> V -> SP1 -> V -> SP0 )

module DF_0dr (C, D_1, D_0, Q_1, Q_0, RN);
  input C;
  input D_1, D_0;
  output Q_1, Q_0;
  input RN;
  wire w_1, w_0;
  wire SP;
  
  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));
  spctrl inst_SP (.C(C), .RN(RN), .Q(SP));
  spalt_ndr inst_Q (.C(C), .D_1(w_1), .D_0(w_0), .Q_1(Q_1), .Q_0(Q_0), .SP(SP));
endmodule


// DF

module DF1_0dr (C, D_1, D_0, Q_1, Q_0); // !!! constant all-zerous spacer
  input C;
  input D_1, D_0;
  output Q_1, Q_0;
  
  delay inst_C_buf1 (.A(C), .Q(C_buf));

  delay inst_D_1_buf (.A(D_1), .Q(D_1_buf));
  delay inst_D_0_buf (.A(D_0), .Q(D_0_buf));

  OAI2111 inst_M_1 (.D(M_0), .C(C), .B(C_buf), .A(D_1_buf), .Q(M_1));
  OAI2111 inst_M_0 (.D(M_1), .C(C), .B(C_buf), .A(D_0_buf), .Q(M_0));

  delay inst_M_1_buf (.A(M_1), .Q(M_1_buf));
  delay inst_M_0_buf (.A(M_0), .Q(M_0_buf));

  AOI2111 inst_Q_1 (.D(Q_0), .C(C), .B(C_buf), .A(M_1_buf), .Q(Q_1));
  AOI2111 inst_Q_0 (.D(Q_1), .C(C), .B(C_buf), .A(M_0_buf), .Q(Q_0));
endmodule


module DF3_0dr (C, D_1, D_0, Q_1, Q_0); // !!! constant all-zerous spacer
  input C;
  input D_1, D_0;
  output Q_1, Q_0;
  wire w_1, w_0;

  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));

  BUF3 inst_Q_1 (.A(w_1), .Q(Q_1));
  BUF3 inst_Q_0 (.A(w_0), .Q(Q_0));
endmodule


// DFC

module DFC1_0dr (C, D_1, D_0, Q_1, Q_0, RN);
  input C;
  input D_1, D_0;
  input RN;
  output Q_1, Q_0;
  wire R;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0), .RN(RN));

  INV1 inst_R (.A(RN), .Q(R));

  NAND21 inst_not_Q_1 (.A(RN), .B(w_1), .Q(not_Q_1));
  NOR21 inst_not_Q_0 (.A(R), .B(w_0), .Q(not_Q_0));

  INV1 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV1 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


module DFC3_0dr (C, D_1, D_0, Q_1, Q_0, RN);
  input C;
  input D_1, D_0;
  input RN;
  output Q_1, Q_0;
  wire R;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0), .RN(RN));

  INV1 inst_R (.A(RN), .Q(R));

  NAND21 inst_not_Q_1 (.A(RN), .B(w_1), .Q(not_Q_1));
  NOR21 inst_not_Q_0 (.A(R), .B(w_0), .Q(not_Q_0));

  INV3 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV3 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


// DFCP

module DFCP1_0dr (C, D_1, D_0, Q_1, Q_0, RN, SN);
  input C;
  input D_1, D_0;
  input RN, SN;
  output Q_1, Q_0;
  wire R, S;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;
  wire not_reset;

  INV1 inst_R (.A(RN), .Q(R));
  INV1 inst_S (.A(SN), .Q(S));

  NAND21 inst_not_reset (.A(R), .B(S), .Q(not_reset));

  DF_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0), .RN(not_reset));

  OAI211 inst_not_Q_1 (.A(w_1), .B(S), .C(RN), .Q(not_Q_1));
  OAI211 inst_not_Q_0 (.A(w_0), .B(R), .C(SN), .Q(not_Q_0));

  INV1 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV1 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


module DFCP3_0dr (C, D_1, D_0, Q_1, Q_0, RN, SN);
  input C;
  input D_1, D_0;
  input RN, SN;
  output Q_1, Q_0;
  wire R, S;
  wire w_1, w_0;
  wire x_1, x_0;
  wire not_reset;

  INV1 inst_R (.A(RN), .Q(R));
  INV1 inst_S (.A(SN), .Q(S));

  NAND21 inst_not_reset (.A(R), .B(S), .Q(not_reset));

  DF_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0), .RN(not_reset));

  OAI211 inst_not_Q_1 (.A(w_1), .B(S), .C(RN), .Q(not_Q_1));
  OAI211 inst_not_Q_0 (.A(w_0), .B(R), .C(SN), .Q(not_Q_0));

  INV3 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV3 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


// DFP

module DFP1_0dr (C, D_1, D_0, Q_1, Q_0, SN);
  input C;
  input D_1, D_0;
  input SN;
  output Q_1, Q_0;
  wire S;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0), .RN(SN));

  INV1 inst_S (.A(SN), .Q(S));

  NOR21 inst_not_Q_1 (.A(S), .B(w_1), .Q(not_Q_1));
  NAND21 inst_not_Q_0 (.A(SN), .B(w_0), .Q(not_Q_0));
  
  INV1 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV1 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


module DFP3_0dr (C, D_1, D_0, Q_1, Q_0, SN);
  input C;
  input D_1, D_0;
  input SN;
  output Q_1, Q_0;
  wire S;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0), .RN(SN));

  INV1 inst_S (.A(SN), .Q(S));

  NOR21 inst_not_Q_1 (.A(S), .B(w_1), .Q(not_Q_1));
  NAND21 inst_not_Q_0 (.A(SN), .B(w_0), .Q(not_Q_0));
  
  INV3 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV3 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


// Tie-Down to Logic Low-Level

module TIE0_0dr (C, Q_1, Q_0, RN);
  input C;
  output Q_1, Q_0;
  input RN;
  wire w_1, w_0;
  wire SP;

  TIE0 inst_w_1 (w_1);
  INV1 inst_w_0 (C, w_0);
  
  spctrl inst_SP (.C(C), .RN(RN), .Q(SP));
  spalt_ndr inst_Q (.C(C), .D_1(w_1), .D_0(w_0), .Q_1(Q_1), .Q_0(Q_0), .SP(SP));
endmodule


module TIE0_1dr (C, Q_1, Q_0, RN);
  input C;
  output Q_1, Q_0;
  input RN;
  wire w_1, w_0;

  TIE0_0dr inst_w (.C(C), .Q_1(w_1), .Q_0(w_0), .RN(RN));
  spinv_ndr inst_Q (.i_1(w_1), .i_0(w_0), .z_1(Q_1), .z_0(Q_0));
endmodule


// Tie-Up to Logic High-Level

module TIE1_0dr (C, Q_1, Q_0, RN);
  input C;
  output Q_1, Q_0;
  input RN;
  wire w_1, w_0;
  wire SP;

  INV1 inst_w_1 (C, w_1);
  TIE0 inst_w_0 (w_0);
  
  spctrl inst_SP (.C(C), .RN(RN), .Q(SP));
  spalt_ndr inst_Q (.C(C), .D_1(w_1), .D_0(w_0), .Q_1(Q_1), .Q_0(Q_0), .SP(SP));
endmodule


module TIE1_1dr (C, Q_1, Q_0, RN);
  input C;
  output Q_1, Q_0;
  input RN;
  wire w_1, w_0;

  TIE1_0dr inst_w (.C(C), .Q_1(w_1), .Q_0(w_0), .RN(RN));
  spinv_ndr inst_Q (.i_1(w_1), .i_0(w_0), .z_1(Q_1), .z_0(Q_0));
endmodule
