//
// This file containes dual-rail flip-flops implementations built on
// timing assumptions. These flip-flop enforces all-zerous spacer on 
// positive phase of clock.
//


// Delay for flip-flops with timing assumptions

module delay (A, Q);
  input A;
  output Q;

  BUF2 inst_q (.A(A), .Q(Q));
endmodule


// DF

module DF1_0dr (C, D_1, D_0, Q_1, Q_0);
  input C;
  input D_1, D_0;
  output Q_1, Q_0;
  wire M_1, M_0;
  wire M_1_buf, M_0_buf;
  wire C_buf;
  
  delay inst_C_buf1 (.A(C), .Q(C_buf));

  delay inst_D_1_buf (.A(D_1), .Q(D_1_buf));
  delay inst_D_0_buf (.A(D_0), .Q(D_0_buf));

  OAI2111 inst_M_1 (.D(M_0), .C(C), .B(C_buf), .A(D_1_buf), .Q(M_1));
  OAI2111 inst_M_0 (.D(M_1), .C(C), .B(C_buf), .A(D_0_buf), .Q(M_0));

  delay inst_M_1_buf (.A(M_1), .Q(M_1_buf));
  delay inst_M_0_buf (.A(M_0), .Q(M_0_buf));

  AOI2111 inst_Q_1 (.D(Q_0), .C(C), .B(C_buf), .A(M_1_buf), .Q(Q_1));
  AOI2111 inst_Q_0 (.D(Q_1), .C(C), .B(C_buf), .A(M_0_buf), .Q(Q_0));
endmodule


module DF3_0dr (C, D_1, D_0, Q_1, Q_0);
  input C;
  input D_1, D_0;
  output Q_1, Q_0;
  wire w_1, w_0;

  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));

  BU3 inst_Q_1 (.A(w_1), .Q(Q_1));
  BU3 inst_Q_0 (.A(w_0), .Q(Q_0));
endmodule


// DFC

module DFC1_0dr (C, D_1, D_0, Q_1, Q_0, RN);
  input C;
  input D_1, D_0;
  input RN;
  output Q_1, Q_0;
  wire R;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));

  INV1 inst_R (.A(RN), .Q(R));

  NAND21 inst_not_Q_1 (.A(RN), .B(w_1), .Q(not_Q_1));
  NOR21 inst_not_Q_0 (.A(R), .B(w_0), .Q(not_Q_0));

  INV1 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV1 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


module DFC3_0dr (C, D_1, D_0, Q_1, Q_0, RN);
  input C;
  input D_1, D_0;
  input RN;
  output Q_1, Q_0;
  wire R;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));

  INV1 inst_R (.A(RN), .Q(R));

  NAND21 inst_not_Q_1 (.A(RN), .B(w_1), .Q(not_Q_1));
  NOR21 inst_not_Q_0 (.A(R), .B(w_0), .Q(not_Q_0));

  INV3 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV3 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


// DFCP

module DFCP1_0dr (C, D_1, D_0, Q_1, Q_0, RN, SN);
  input C;
  input D_1, D_0;
  input RN, SN;
  output Q_1, Q_0;
  wire R, S;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));

  INV1 inst_R (.A(RN), .Q(R));
  INV1 inst_S (.A(SN), .Q(S));

  OAI211 inst_not_Q_1 (.A(w_1), .B(S), .C(RN), .Q(not_Q_1));
  OAI211 inst_not_Q_0 (.A(w_0), .B(R), .C(SN), .Q(not_Q_0));

  INV1 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV1 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


module DFCP3_0dr (C, D_1, D_0, Q_1, Q_0, RN, SN);
  input C;
  input D_1, D_0;
  input RN, SN;
  output Q_1, Q_0;
  wire R, S;
  wire w_1, w_0;
  wire x_1, x_0;

  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));

  INV1 inst_R (.A(RN), .Q(R));
  INV1 inst_S (.A(SN), .Q(S));

  OAI211 inst_not_Q_1 (.A(w_1), .B(S), .C(RN), .Q(not_Q_1));
  OAI211 inst_not_Q_0 (.A(w_0), .B(R), .C(SN), .Q(not_Q_0));

  INV3 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV3 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


// DFP

module DFP1_0dr (C, D_1, D_0, Q_1, Q_0, SN);
  input C;
  input D_1, D_0;
  input SN;
  output Q_1, Q_0;
  wire S;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));

  INV1 inst_S (.A(SN), .Q(S));

  NOR21 inst_not_Q_1 (.A(S), .B(w_1), .Q(not_Q_1));
  NAND21 inst_not_Q_0 (.A(SN), .B(w_0), .Q(not_Q_0));
  
  INV1 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV1 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


module DFP3_0dr (C, D_1, D_0, Q_1, Q_0, SN);
  input C;
  input D_1, D_0;
  input SN;
  output Q_1, Q_0;
  wire S;
  wire w_1, w_0;
  wire not_Q_1, not_Q_0;

  DF1_0dr inst_w (.C(C), .D_1(D_1), .D_0(D_0), .Q_1(w_1), .Q_0(w_0));

  INV1 inst_S (.A(SN), .Q(S));

  NOR21 inst_not_Q_1 (.A(S), .B(w_1), .Q(not_Q_1));
  NAND21 inst_not_Q_0 (.A(SN), .B(w_0), .Q(not_Q_0));
  
  INV3 inst_Q_1 (.A(not_Q_1), .Q(Q_1));
  INV3 inst_Q_0 (.A(not_Q_0), .Q(Q_0));
endmodule


// Tie-Down to Logic Low-Level

module TIE0_0dr (C, Q_1, Q_0);
  input C;
  output Q_1, Q_0;

  TIE0 inst_Q_1 (Q_1);
  INV1 inst_Q_0 (C, Q_0);
endmodule


module TIE0_1dr (C, Q_1, Q_0);
  input C;
  output Q_1, Q_0;

  BUF2 inst_Q_1 (C, Q_1);
  TIE1 inst_Q_0 (Q_0);
endmodule


// Tie-Up to Logic High-Level

module TIE1_0dr (C, Q_1, Q_0);
  input C;
  output Q_1, Q_0;

  INV1 inst_Q_1 (C, Q_1);
  TIE0 inst_Q_0 (Q_0);
endmodule


module TIE1_1dr (C, Q_1, Q_0);
  input C;
  output Q_1, Q_0;

  TIE1 inst_Q_1 (Q_1);
  BUF2 inst_Q_0 (C, Q_0);
endmodule
